"""

Usage:
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> 
                [--age=<age>][--share=<share>][--long][--time|--size][--reverse][--human-readable]
   iguard_ls --hostname=<hostname> [--username=<username>] [--password=<password>] --dde=<dde_id> --list
   iguard_ls --hostname=<hostname> --username=<username> --password=<password> --save
   iguard_ls --delete
   iguard_ls --help
   iguard_ls --version


Options:
   -H hostname --hostname=<hostname>   DNS name/IP  of InfiniGuard
   -u username --username=<username>   Administrator's username
   -p password --password=<password>   Administrator's password
   -d dde_id, --dde=<dde_id>           DDE ID (1/2)
   -s share, --share=<share>           Display only share name <share>
   -a age, --age=<age>                 Only display files older than <age> days
   -D, --delete                        Delete all stored credentials
   -S, --save                          Save credentials for hostname
   -L, --list                          List all share names
   -l, --long                          Use long display format
   -t, --time                          Sort by time
   --size                              Sort by size
   -r, --reverse                       Sort in reverse order
   -h, --human-readable                Display size in human readable format
   --help                              Show this screen
   -V, --version                       Show iguard_ls version
"""
from __future__ import print_function
from .logsetup import UserLogger, setup_logging, has_warnings
from docopt import docopt
from .__version__ import __version__ as iguard_ls_version
import json
import os
from infinisdk import InfiniBox, Q
import requests
import ssl
import sys
import base64
import lzma
import arrow
from time import ctime, mktime, gmtime, asctime, time
import stat
from infinisdk.core.exceptions import APICommandFailed
from infi.credentials_store import CLICredentialsStore, Credentials
import pickle
from time import sleep
import traceback

requests.packages.urllib3.disable_warnings()

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context


log = UserLogger(__name__)


def get_function_name():
    return traceback.extract_stack(None, 2)[0][2]


def load_cookie(infinibox, username, hostname):
    try:
        with open('/tmp/iguard_ls_cookie_{}_{}'.format(username, hostname), 'rb') as f:
            return pickle.load(f)
    except Exception:
        log.debug("Cookie not found.")
    return None


def save_cookie(infinibox, username, hostname):
    cookie = None
    try:
        with open('/tmp/iguard_ls_cookie_{}_{}'.format(username, hostname), 'wb') as f:
            pickle.dump(infinibox.api._session.cookies.copy(), f)
    except Exception as e:
        log.debug("Failed to save cookie")
        log.debug(repr(e))


def my_login(infinibox, username, hostname):
    cookie = load_cookie(infinibox, username, hostname)
    if not cookie:
        log.debug("Cookie not found. Logging in.")
        infinibox.login()
        cookie = save_cookie(infinibox, username, hostname)
    else:
        log.debug("Loaded saved cookie")
        infinibox.api._session.cookies = cookie
        try:
            infinibox.get_version()
            # It didn't use the cookie, but looged in
            if infinibox.is_logged_in():
                log.debug("Logged in, cookie expired. Saving.")
                save_cookie(infinibox, username, hostname)
        except Exception:
            log.debug("Cookie not accepted")
            infinibox.api._session.cookies.clear()
            infinibox.login()
            save_cookie(infinibox, username, hostname)


class MyCLICrdentialsStore(CLICredentialsStore):
    def _get_file_folder(self):
        return ".infi.credentials_store"

    def authenticate(self, key, credentials):
        return True

    def ask_credentials_prompt(self, key):
        print('Connecting to system ' + str(key))


def delete_credentials(store):
    print("Deleting all credentials")
    store.reset_credentials()


def save_credentials(store, hostname, username, password):
    credentials = Credentials(username, password)
    store.set_credentials(hostname, credentials)
    print("Stored credentials for host {}".format(hostname))


def validate_config(arguments, store):
    config = {}
    if arguments['--help']:
        docopt(__doc__, '-h')

    for key, value in arguments.items():
        config[key[2:]] = value
    if not config['delete']:
        if not config['username'] or not config['password']:
            credentials = store.get_credentials(config['hostname'])
            config['username'] = credentials.get_username()
            config['password'] = credentials.get_password()
    if not config['delete'] and not config['save']:
        try:
            dde_id = int(config['dde'])
            if dde_id not in [1, 2]:
                raise Exception
            config['dde'] = dde_id
        except Exception as e:
            raise Exception('DDE ID should be 1 or 2. Exiting.')
    return config


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_err(*args, **kwargs):
    print(f'{bcolors.FAIL}{" ".join([a for a in args])}{bcolors.ENDC}', file=sys.stderr, **kwargs)


def _print_files(share_data, app_share, is_long=False, human_readable=False):
    def human_readable_size(size, decimal_places=1):
        for unit in ['', 'K', 'M', 'G', 'T', 'P']:
            if size < 1024.0:
                break
            size /= 1024.0
        if not unit:
            return "{}".format(size)
        else:
            return "{size:.{decimal_places}f}{unit}".format(size=size, decimal_places=decimal_places, unit=unit)

    for file in share_data:
        path = file.get('filename')
        if is_long:
            size = file.get('size')
            mtime = asctime(gmtime(file.get('mtime')))
            mode = stat.filemode(file.get('mode'))
            uid = file.get('uid')
            gid = file.get('gid')
            if human_readable:
                size_format = '{:>7}'.format(human_readable_size(size))
            else:
                size_format = '{:<12}'.format(size)
            print('{} {:<4} {:<4} {} {} {}/{}'.format(mode,
                                                      uid, gid, mtime, size_format, app_share, path))
        else:
            print('{}/{}'.format(app_share, path))


def get_shares(config):
    infinibox = config.get('infinibox')
    dde_id = config.get('dde')
    get = infinibox.api.get(
        "/api/dde/{}/api/nas/app_specific/shares/".format(dde_id))
    res = get.get_result()
    return [a.get('name') for a in res]


def print_shares(config):
    for share in get_shares(config):
        print(share)


def is_task_finished(infinibox, dde_id, task_id):
    task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
    get = infinibox.api.get(task_url)
    status = get.get_result().get('status')
    return status != 'RUNNING'


def get_files(infinibox, dde_id, task_id, last_id):
    result_url = "/api/dde/{}/api/asynctasks/{}/file_results?sort=id&page_size=1000&page=1&id=gt:{}".format(
        dde_id, task_id, last_id)
    get = infinibox.api.get(result_url)
    res = get.get_result()
    return res


def print_files(config):
    log.info('Running {}'.format(get_function_name()))

    dde_id = config['dde']
    infinibox = config['infinibox']
    EXCLUDES = ['infinibox', 'version']
    data = {k: v for k, v in config.items(
    ) if k not in EXCLUDES and v is not None}
    log.debug(data)
    is_long = config.get('long')
    is_human_readable = config.get('human-readable')
    is_time = config.get('time')
    is_size = config.get('size')
    is_reverse = True if config.get('reverse') else False
    is_sorted = is_time or is_size
    share = config.get('share', None)
    shares = get_shares(config)
    if share:
        if share in shares:
            shares = [share]
        else:
            print('Share {} not found on DDE. Exiting.'.format(share))
            sys.exit(1)
    age = config.get('age', 0)
    if not age:
        age_query = ''
    else:
        now = int(time())
        seconds = int(age) * 3600 * 24
        age_query = '?mtime=lte:{}'.format(now - seconds)

    log.info("Shares: {}".format(shares))
    rc = 0
    for share in shares:
        post = infinibox.api.post(
            "/api/dde/{}/api/nas/app_specific/shares/{}/list_files{}".format(dde_id, share, age_query))

        task_id = post.get_result().get('task_id')
        task_url = "/api/dde/{}/api/asynctasks/{}".format(dde_id, task_id)
        log.debug(task_url)
        result = []
        try:
            last_id = 0
            while True:
                finished = is_task_finished(infinibox, dde_id, task_id)
                out = get_files(infinibox, dde_id, task_id, last_id)
                if out:
                    last_id = out[-1].get('id')
                    if not is_sorted:
                        _print_files(out, share, is_long, is_human_readable)
                    else:
                        result += out
                if finished and not out:
                    break
                else:
                    sleep(0.1)

            if is_sorted:
                if is_size:
                    sortfield = 'size'
                else:
                    sortfield = 'mtime'

                out = sorted(result, key=lambda k: k[sortfield], reverse=not is_reverse)
                _print_files(out, share, is_long, is_human_readable)

        finally:
            delete = infinibox.api.delete(task_url)
            log.debug('Delete status code: {}'.format(delete.status_code))

    sys.exit(rc)


def infiniguard_ls():
    try:
        log_handler = setup_logging()
        with log_handler.applicationbound():
            arguments = docopt(__doc__, version=iguard_ls_version)
            store = MyCLICrdentialsStore("iguard_ls")
            config = validate_config(arguments, store)
            hostname = config['hostname']
            username = config['username']
            password = config['password']
            dde = config['dde']
            if not config['save'] and not config['delete']:
                infinibox = InfiniBox(hostname, use_ssl=True,
                                      auth=(username, password))
                my_login(infinibox, username, hostname)
                config['infinibox'] = infinibox
                get = infinibox.api.get(
                    '/api/dde/{}/api/asynctasks/'.format(dde))
                if not get.response.content:
                    print_err('This system does not support iguard_ls.')
                    sys.exit(1)

            if config['delete']:
                delete_credentials(store)
            elif config['save']:
                save_credentials(store, hostname, username, password)
            elif config['list']:
                print_shares(config)
            else:
                print_files(config)

    except APICommandFailed as e:
        if e.status_code == 404:
            print_err('This system does not support iguard_ls')
        elif e.status_code == 503:
            print_err('System Unreachable')
        else:
            print_err(f'ERROR: {e.message}')
        sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)
